#training codes of the model in ablation study 2^{nd}-image+feature and 3^{rd}-L2+NL2. 
#please train the Wide-ResNet using 'Train_CIFAR100.py' and save ckpts models in'ckpts_w_100_1' before running this script.

import tensorflow as tf
from tflearn.layers.conv import global_avg_pool
from tensorflow.contrib.layers import batch_norm, flatten
from tensorflow.contrib.layers import xavier_initializer
from tensorflow.contrib.framework import arg_scope
import random

physical_devices = tf.config.experimental.list_physical_devices('GPU') 
assert len(physical_devices) > 0, "Not enough GPU hardware devices available"
tf.config.experimental.set_memory_growth(physical_devices[0], True)



def image_augmentor(image,flip_prob=0.5,crop_prob = 0.5):
    
    if flip_prob is not None:
        flip_lr_prob = tf.random_uniform([], 0., 1.)
        
        image = tf.cond(
            tf.less(flip_lr_prob, flip_prob),
            lambda: tf.reverse(image, [1]),
            lambda: image
        )
    
    image = (image - tf.constant([129.30416560546874,124.0699626953125,112.43405005859375]))/tf.constant([68.17024289920643,65.39180804385748,70.41837018804927])
    
    return image
    
def data_augment_in_gpu(images):
    images_list = tf.unstack(images,axis=0)
    output_images = []
    for idx in range(len(images_list)):
        images = image_augmentor(image = images_list[idx],
                flip_prob=0.5)    
        output_images.append(images)
    images = tf.stack(output_images,axis=0)
    return images
        
def parser(record):
    features = tf.parse_single_example(
        record,
        features={
            'image':tf.FixedLenFeature([],tf.string),
            'label':tf.FixedLenFeature([],tf.int64)
        })
    
    images = tf.decode_raw(features['image'], tf.uint8) 
    images = tf.reshape(images,[32,32,3])
    images = tf.cast(images, tf.float32)
    labels = tf.cast(features['label'],tf.int32)

    return images, labels

def wide_basic(inputs, in_planes, out_planes, dropout_rate, stride, name, index,is_train):
    with tf.variable_scope('wide_basic' + name):
        if stride != 1 or in_planes != out_planes:
            skip_c = tf.layers.conv2d(inputs, out_planes, kernel_size=1, strides=stride, use_bias = True, padding='SAME')
        else:
            skip_c = inputs
        
        x = tf.contrib.layers.batch_norm(inputs, updates_collections=None, decay=0.9, scale=True, center=True, is_training = is_train)
        x = tf.nn.relu(x)
        x = tf.layers.conv2d(x, out_planes, kernel_size=3, strides=1, use_bias = True, padding='SAME')
        x = tf.layers.dropout(x, rate=0.1, training=is_train)
        x = tf.contrib.layers.batch_norm(x, updates_collections=None, decay=0.9, scale=True, center=True, is_training = is_train)
        x = tf.nn.relu(x)
        x = tf.layers.conv2d(x, out_planes, kernel_size=3, strides=stride, use_bias = True, padding='SAME')
        
        x = tf.add(skip_c, x)
        
    return x

def wide_layer(out, in_planes, out_planes, num_blocks, dropout_rate, stride, name,is_train):
    with tf.variable_scope('wide_layer' + name):
        strides = [stride] + [1]*int(num_blocks-1)
        i = 0
        for strid in strides:
            out = wide_basic(out, in_planes, out_planes, dropout_rate, strid, name = 'layer1_'+str(i)+'_', index = i,is_train = is_train)
            in_planes = out_planes
            i += 1
        
    return out
        
def make_resnet_filter(ins, depth=28, widen_factor=10, dropout_rate = 0.3, reuse=False,is_train=True):
    n = (depth-4)/6
    k = widen_factor
    print('| Wide-Resnet %dx%d' %(depth, k))
    nStages = [16, 16*k, 32*k, 64*k]
    with tf.variable_scope('WRN', reuse=reuse):
        x = tf.layers.conv2d(ins, nStages[0], kernel_size=3, strides=1, use_bias = True, padding='SAME')
        x = wide_layer(x, nStages[0], nStages[1], n, dropout_rate, stride=1, name = 'layer1_',is_train=is_train)
        x = wide_layer(x, nStages[1], nStages[2], n, dropout_rate, stride=2, name = 'layer2_',is_train=is_train)
        x = wide_layer(x, nStages[2], nStages[3], n, dropout_rate, stride=2, name = 'layer3_',is_train=is_train)
        x = tf.contrib.layers.batch_norm(x, updates_collections=None, decay=0.9, scale=True, center=True, is_training = is_train)
        x = tf.nn.relu(x)
        x = tf.keras.layers.AvgPool2D([8,8])(x)
        x = tf.reshape(x, (-1, 640))
        low = x
        x = tf.layers.dense(x, num_classes,use_bias = False)

        return low,x

def mask_fc(mask_type, in_channels, out_channels, code_length):
    
    mask = np.zeros((out_channels*code_length,in_channels*code_length))
    
    for f in range(0 if mask_type == 1 else 1, code_length):
        start_row = f*out_channels
        end_row = (f+1)*out_channels
        
        start_col = 0
        end_col = f*in_channels if mask_type == 0 else (f+1)*in_channels
        if start_col != end_col:
            mask[start_row:end_row, start_col:end_col] = 1
            
    return mask.transpose()

def MaskedFullyConnection(x,in_channels,out_channels,code_length,mask_type,name):
    kernel = tf.get_variable(name=name,dtype=tf.float32,shape=[code_length*in_channels,code_length*out_channels],initializer=tf.contrib.layers.xavier_initializer(uniform=False))
    
    x = tf.reshape(x,[batch_size,code_length*in_channels])
    mask = tf.py_func(mask_fc,[mask_type,in_channels,out_channels,code_length],[tf.double])
    output = tf.matmul(x,kernel*mask)
    
    return output

def ILCE(z,name):
    with tf.variable_scope(name):
        
        z = tf.reshape(z,[batch_size,1,1,640])
        z = tf.layers.conv2d(z, num_classes, kernel_size=1, strides=1, padding='SAME')
        z = tf.nn.relu(z)
        z = tf.layers.conv2d(z, num_classes, kernel_size=1, strides=1, padding='SAME')
        z = tf.squeeze(z)
        z = tf.nn.sigmoid(z)
        
        z_dist = MaskedFullyConnection(z,1,32,100,0,'kernel_0')
        z_dist = tf.nn.leaky_relu(z_dist)
        z_dist = MaskedFullyConnection(z_dist,32,32,100,1,'kernel_1')
        z_dist = tf.nn.leaky_relu(z_dist)
        z_dist = MaskedFullyConnection(z_dist,32,32,100,1,'kernel_2')
        z_dist = tf.nn.leaky_relu(z_dist)
        z_dist = MaskedFullyConnection(z_dist,32,32,100,1,'kernel_3')
        z_dist = tf.nn.leaky_relu(z_dist)
        z_dist = MaskedFullyConnection(z_dist,32,100,100,1,'kernel_4')
    
        x = tf.reshape(z,[batch_size,1,1,num_classes])
        x = tf.layers.conv2d(x, 640, kernel_size=1, strides=1, padding='SAME')
        x = tf.nn.swish(x)
        x = tf.layers.conv2d(x, 640, kernel_size=1, strides=1, padding='SAME')
        x = tf.nn.swish(x)
        x = tf.layers.conv2d(x, 640, kernel_size=1, strides=1, padding='SAME')
      
        return x,z,z_dist
    
def theround(x):
    index = np.clip(x.astype(np.int32),0,99)
    matrix = np.zeros((batch_size,100,100))
    
    for i in range(batch_size):
        for j in range(100):
            matrix[i,j,index[i,j]] = 1
    
    return matrix.astype(np.float32)

    
tf.reset_default_graph()

pretrained_model = tf.train.latest_checkpoint('ckpts_w_100_1')
logs_train_dir = 'ckpts_w_100_2_auto'
data_train_tfr = 'TFRecord_trainset_100'
total_train_sample = 50000
num_classes = 100

init_learning_rate = 1e-4
epsilon = 1e-6 
batch_size = 128
is_train = True
total_epochs = 300
shuffle_buffer = 1000
steps_per_epoch = int(total_train_sample/batch_size)

global_step = tf.Variable(0, name='global_step', trainable=False)    
dataset = tf.data.TFRecordDataset(data_train_tfr)
dataset = dataset.map(parser)
dataset = dataset.repeat(total_epochs)
dataset = dataset.shuffle(shuffle_buffer).batch(batch_size,drop_remainder=True)
iterator = dataset.make_initializable_iterator()
image_batch, label_batch = iterator.get_next()
image_batch = data_augment_in_gpu(image_batch)
label_batch_oh = tf.one_hot(label_batch,num_classes)

lr = tf.cond(global_step < steps_per_epoch * int(total_epochs*0.5),lambda: tf.constant(init_learning_rate),lambda: tf.cond(global_step < steps_per_epoch * int(total_epochs*0.75),lambda: tf.constant(init_learning_rate/10),lambda: tf.constant(init_learning_rate/100)))

image_batch = tf.reshape(image_batch,[-1,32,32,3])
low,low_1 = make_resnet_filter(image_batch, depth=28, widen_factor=10, dropout_rate = 0.0, reuse=False,is_train=True)
rec,z,z_dist = ILCE(low,'ILCE')

z_dist = tf.log(tf.clip_by_value(tf.nn.softmax(tf.reshape(z_dist,[batch_size,100,100]), dim=-1),0.001,0.999))
z = tf.reshape(z,[batch_size,100])
matrix = tf.py_func(theround,[z*100],[tf.float32])[0]

selected = matrix*z_dist 
selected = tf.squeeze(tf.reduce_sum(selected, axis=-1))
loss_0 = -tf.reduce_mean(tf.reduce_sum(selected, axis=-1))

rec = tf.reshape(rec,[batch_size,-1])
low = tf.reshape(low,[batch_size,-1])

loss_1 = tf.reduce_mean(tf.reduce_sum((low - rec)**2,-1))

loss = 0.1*loss_0 + loss_1

g_list = tf.global_variables()
exclude = ['ILCE']
exclude += ['global_step']
exclude += ['is_training']
exclude += ['adam']
variables_to_restore = tf.contrib.slim.get_variables_to_restore(
        exclude=exclude)
variables_map = {}
for v in variables_to_restore:
    variables_map[v.name.split(':')[0]] = v
saver_re = tf.train.Saver(variables_map)

saver = tf.train.Saver(max_to_keep=5)

cls_var = [g for g in tf.global_variables() if ('ILCE' in g.name)]
optimizer = tf.train.AdamOptimizer(learning_rate = lr,name='adam')
train = optimizer.minimize(loss,global_step=global_step,var_list = cls_var)

correct = tf.nn.in_top_k(tf.squeeze(low_1), label_batch, 1)
correct = tf.cast(correct, tf.float16)
accuracy = tf.reduce_mean(correct)

for each in os.listdir(logs_train_dir):
    if each != '.ipynb_checkpoints':
        os.remove(logs_train_dir+ '/' + each)

with tf.Session() as sess:  
    sess.run((tf.global_variables_initializer(),
              tf.local_variables_initializer()))
    sess.run(iterator.initializer)
    saver_re.restore(sess,pretrained_model)
            
    count = 0
    while True:
        try:
            _,tra_loss_0,tra_loss_1,this_acc = sess.run([train,loss_0,loss_1,accuracy])
            if count %100 == 0:
                print('step: %d,loss_0:%f,loss_1:%f,acc:%f'%(count,tra_loss_0,tra_loss_1,this_acc))
            if count % int(steps_per_epoch*5) == 0 and count >= 0.5*total_epochs*steps_per_epoch:      
                checkpoint_path = os.path.join(logs_train_dir, 'model.ckpt')
                saver.save(sess, checkpoint_path, global_step=count)
            count += 1
        except tf.errors.OutOfRangeError:
            print('Done!')
            break