#Code for making our training/validation/test TFrecord data of CIFAR-10. 
#Please download CIFAR-10 as 'cifar-10-batches-py' before running this script. 


from __future__ import print_function
from six.moves import cPickle as pickle
import numpy as np
import os
import platform
import matplotlib.pyplot as plt
import tensorflow as tf
import sys


def _int64_feature(value):
    return tf.train.Feature(int64_list = tf.train.Int64List(value=[value]))

def bytes_feature(values):
    if not isinstance(values, (tuple, list)):
        values = [values]
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=values))

def load_pickle(f):
    version = platform.python_version_tuple() 
    if version[0] == '2':
        return  pickle.load(f) 
    elif version[0] == '3':
        return  pickle.load(f, encoding='latin1')
    raise ValueError("invalid python version: {}".format(version))

def load_CIFAR_batch(filename):
  with open(filename, 'rb') as f:
    datadict = load_pickle(f) 
    X = datadict['data']      
    Y = datadict['labels']  
    
    X = X.reshape(10000, 3, 32, 32).transpose(0,2,3,1).astype("float")
    Y = np.array(Y)
    return X, Y

def load_CIFAR10(ROOT):
  xs = []
  ys = []
  
  for b in range(1,6):
    f = os.path.join(ROOT, 'data_batch_%d' % (b, ))
    X, Y = load_CIFAR_batch(f)
    xs.append(X)
    ys.append(Y)    
  Xtr = np.concatenate(xs) 
  Ytr = np.concatenate(ys)
  del X, Y

  Xte, Yte = load_CIFAR_batch(os.path.join(ROOT, 'test_batch'))
  return Xtr, Ytr, Xte, Yte

ROOT = 'cifar-10-batches-py'
X, Y, Xte, Yte = load_CIFAR10(ROOT)

index_tr=np.arange(X.shape[0])
np.random.shuffle(index_tr)

Xval = X[index_tr[X.shape[0]-2000:]]
Yval = Y[index_tr[X.shape[0]-2000:]]

Xtr = X[index_tr[:X.shape[0]-2000]]
Ytr = Y[index_tr[:X.shape[0]-2000]]

filename_train = 'TFRecord_trainset_10'
filename_val = 'TFRecord_valset_10'
filename_test = 'TFRecord_testset_10'

writer_train = tf.python_io.TFRecordWriter(filename_train)
writer_val = tf.python_io.TFRecordWriter(filename_val)
writer_test = tf.python_io.TFRecordWriter(filename_test)

count = 0
for each in range(Xtr.shape[0]):
    img = np.uint8(Xtr[each,:,:,:].squeeze())
    label = Ytr[each]
            
    features = {
                'label':_int64_feature(label),
                'image':bytes_feature(img.tobytes())
                    }

    example = tf.train.Example(features=tf.train.Features(feature=features))
    writer_train.write(example.SerializeToString())
    sys.stdout.flush()
    count += 1
    
print(count)
writer_train.close()

count = 0
for each in range(Xval.shape[0]):
    img = np.uint8(Xval[each,:,:,:].squeeze())
    label = Yval[each]
            
    features = {
                'label':_int64_feature(label),
                'image':bytes_feature(img.tobytes())
                    }

    example = tf.train.Example(features=tf.train.Features(feature=features))
    writer_val.write(example.SerializeToString())
    sys.stdout.flush()
    count += 1
    
print(count)
writer_val.close()

count = 0
for each in range(Xte.shape[0]):
    img = np.uint8(Xte[each,:,:,:].squeeze())
    label = Yte[each]
            
    features = {
                'label':_int64_feature(label),
                'image':bytes_feature(img.tobytes())
                    }

    example = tf.train.Example(features=tf.train.Features(feature=features))
    writer_test.write(example.SerializeToString())
    sys.stdout.flush()
    count += 1
print(count)
writer_test.close()