#training codes of layerwise semantic reconstruction.
#please train the Wide-ResNet using 'Train_CIFAR_10.py' and save ckpts models in'ckpts_w_10_1' before running this script.


import tensorflow as tf
from tflearn.layers.conv import global_avg_pool
from tensorflow.contrib.layers import batch_norm, flatten
from tensorflow.contrib.layers import xavier_initializer
from tensorflow.contrib.framework import arg_scope
import random


def image_augmentor(image,flip_prob=0.5):
    
    if flip_prob is not None:
        flip_lr_prob = tf.random_uniform([], 0., 1.)
        
        image = tf.cond(
            tf.less(flip_lr_prob, flip_prob),
            lambda: tf.reverse(image, [1]),
            lambda: image
        )
    
    image = (image - tf.constant([129.30416560546874,124.0699626953125,112.43405005859375]))/tf.constant([68.17024289920643,65.39180804385748,70.41837018804927])
    
    return image
    
def data_augment_in_gpu(images):
    images_list = tf.unstack(images,axis=0)
    output_images = []
    for idx in range(len(images_list)):
        images = image_augmentor(image = images_list[idx],
                flip_prob=0.5)    
        output_images.append(images)
    images = tf.stack(output_images,axis=0)
    return images
        
def parser(record):
    features = tf.parse_single_example(
        record,
        features={
            'image':tf.FixedLenFeature([],tf.string),
            'label':tf.FixedLenFeature([],tf.int64)
        })
    
    images = tf.decode_raw(features['image'], tf.uint8) 
    images = tf.reshape(images,[32,32,3])
    images = tf.cast(images, tf.float32)
    labels = tf.cast(features['label'],tf.int32)

    return images, labels


def wide_basic(inputs, in_planes, out_planes, dropout_rate, stride, name, index,is_train):
    with tf.variable_scope('wide_basic' + name):
        if stride != 1 or in_planes != out_planes:
            skip_c = tf.layers.conv2d(inputs, out_planes, kernel_size=1, strides=stride, use_bias = True, padding='SAME')
        else:
            skip_c = inputs
        
        x = tf.contrib.layers.batch_norm(inputs, updates_collections=None, decay=0.9, scale=True, center=True, is_training = is_train)
        x = tf.nn.relu(x)
        x = tf.layers.conv2d(x, out_planes, kernel_size=3, strides=1, use_bias = True, padding='SAME')
        x = tf.layers.dropout(x, rate=0.1, training=is_train)
        x = tf.contrib.layers.batch_norm(x, updates_collections=None, decay=0.9, scale=True, center=True, is_training = is_train)
        x = tf.nn.relu(x)
        x = tf.layers.conv2d(x, out_planes, kernel_size=3, strides=stride, use_bias = True, padding='SAME')
      
        x = tf.add(skip_c, x)
        
    return x

def wide_layer(out, in_planes, out_planes, num_blocks, dropout_rate, stride, name,is_train):
    with tf.variable_scope('wide_layer' + name):
        strides = [stride] + [1]*int(num_blocks-1)
        i = 0
        for strid in strides:
            out = wide_basic(out, in_planes, out_planes, dropout_rate, strid, name = 'layer1_'+str(i)+'_', index = i,is_train = is_train)
            in_planes = out_planes
            i += 1
        
    return out
        
def make_resnet_filter(ins, depth=28, widen_factor=10, dropout_rate = 0.0, reuse=False,is_train=True):
    n = (depth-4)/6
    k = widen_factor
    print('| Wide-Resnet %dx%d' %(depth, k))
    nStages = [16, 16*k, 32*k, 64*k]
    with tf.variable_scope('WRN', reuse=reuse):
        x = tf.layers.conv2d(ins, nStages[0], kernel_size=3, strides=1, use_bias = True, padding='SAME')
        x = wide_layer(x, nStages[0], nStages[1], n, dropout_rate, stride=1, name = 'layer1_',is_train=is_train)
        x = wide_layer(x, nStages[1], nStages[2], n, dropout_rate, stride=2, name = 'layer2_',is_train=is_train)
        x = wide_layer(x, nStages[2], nStages[3], n, dropout_rate, stride=2, name = 'layer3_',is_train=is_train)
        x = tf.contrib.layers.batch_norm(x, updates_collections=None, decay=0.9, scale=True, center=True, is_training = is_train)
        x = tf.nn.relu(x)
        x = tf.keras.layers.AvgPool2D([8,8])(x)
        x = tf.reshape(x, (-1, 640))
        low = x
        x = tf.layers.dense(x, num_classes,use_bias = False)
        
        return low,x

def ILCE_0(low_1,name):
    with tf.variable_scope(name):
        
        rec_0 = tf.reshape(low_1,[batch_size,1,1,num_classes])
        rec_0 = tf.layers.conv2d(rec_0, 640, kernel_size=1, strides=1, padding='SAME')
        rec_0 = tf.nn.swish(rec_0)
        rec_0 = tf.layers.conv2d(rec_0, 640, kernel_size=1, strides=1, padding='SAME')
        rec_0 = tf.nn.swish(rec_0)
        rec_0 = tf.layers.conv2d(rec_0, 640, kernel_size=1, strides=1, padding='SAME')
      
        return rec_0

def ILCE_1(softmax_prob,name):
    with tf.variable_scope(name):
        
        rec_1 = tf.reshape(softmax_prob,[batch_size,1,1,num_classes])
        rec_1 = tf.layers.conv2d(rec_1, 512, kernel_size=1, strides=1, padding='SAME')
        rec_1 = tf.nn.swish(rec_1)
        rec_1 = tf.layers.conv2d(rec_1, 256, kernel_size=1, strides=1, padding='SAME')
        rec_1 = tf.nn.swish(rec_1)
        rec_1 = tf.layers.conv2d(rec_1, num_classes, kernel_size=1, strides=1, padding='SAME')
        
        return rec_1
    
tf.reset_default_graph()
pretrained_model = tf.train.latest_checkpoint('ckpts_w_10_1')
logs_train_dir = 'ckpts_w_10_2'
data_train_tfr = 'TFRecord_trainset_10'
total_train_sample = 50000
num_classes = 10

init_learning_rate = 1e-4
epsilon = 1e-6 
dropout_rate = 0.
batch_size = 128
is_train = True
total_epochs = 300
shuffle_buffer = 1000
steps_per_epoch = int(total_train_sample/batch_size)
total_steps = steps_per_epoch*total_epochs

global_step = tf.Variable(0, name='global_step', trainable=False)    
dataset = tf.data.TFRecordDataset(data_train_tfr)
dataset = dataset.map(parser)
dataset = dataset.repeat(total_epochs)
dataset = dataset.shuffle(shuffle_buffer).batch(batch_size,drop_remainder=True)
iterator = dataset.make_initializable_iterator()
image_batch, label_batch = iterator.get_next()
image_batch = data_augment_in_gpu(image_batch)
label_batch_oh = tf.one_hot(label_batch,num_classes)

lr = tf.cond(global_step < steps_per_epoch * int(total_epochs*0.5),lambda: tf.constant(init_learning_rate),lambda: tf.cond(global_step < steps_per_epoch * int(total_epochs*0.75),lambda: tf.constant(init_learning_rate/10),lambda: tf.constant(init_learning_rate/100)))

image_batch = tf.reshape(image_batch,[-1,32,32,3])
low_0,low_1 = make_resnet_filter(image_batch, depth=28, widen_factor=10, dropout_rate = dropout_rate, reuse=False, is_train=is_train)

low_1_ = tf.nn.softmax(low_1/100)
loss_2 = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(labels=label_batch_oh, logits=low_1))

rec_0 = ILCE_0(low_1,'ILCE_0')
rec_1 = ILCE_1(low_1_,'ILCE_1')

low_1 = low_1/100
rec_0 = tf.reshape(rec_0,[batch_size,-1])
low_0 = tf.reshape(low_0,[batch_size,-1])
rec_1 = tf.reshape(rec_1,[batch_size,-1])
low_1 = tf.reshape(low_1,[batch_size,-1])

loss_0 = tf.reduce_mean(tf.reduce_sum((low_0 - rec_0)**2,-1))
loss_1 = tf.reduce_mean(tf.reduce_sum((low_1 - rec_1)**2,-1))

loss = loss_0 + loss_1 + loss_2

g_list = tf.global_variables()
exclude = ['ILCE_1']
exclude += ['ILCE_0']
exclude += ['global_step']
exclude += ['is_training']
exclude += ['adam']
variables_to_restore = tf.contrib.slim.get_variables_to_restore(
        exclude=exclude)
variables_map = {}
for v in variables_to_restore:
    variables_map[v.name.split(':')[0]] = v
saver_re = tf.train.Saver(variables_map)

saver = tf.train.Saver(max_to_keep=5)

cls_var = [g for g in tf.global_variables() if ('ILCE_0' in g.name or 'ILCE_1' in g.name or 'dense' in g.name)]
optimizer = tf.train.AdamOptimizer(learning_rate = lr,name='adam')
train = optimizer.minimize(loss,global_step=global_step,var_list = cls_var)

correct = tf.nn.in_top_k(tf.squeeze(low_1), label_batch, 1)
correct = tf.cast(correct, tf.float16)
accuracy = tf.reduce_mean(correct)

for each in os.listdir(logs_train_dir):
    if each != '.ipynb_checkpoints':
        os.remove(logs_train_dir+ '/' + each)

with tf.Session() as sess:  
    sess.run((tf.global_variables_initializer(),
              tf.local_variables_initializer()))
    sess.run(iterator.initializer)
    saver_re.restore(sess,pretrained_model)
            
    count = 0
    while True:
        try:
            _,tra_loss_0,tra_loss_1,this_acc = sess.run([train,loss_0,loss_1,accuracy])
            if count %100 == 0:
                print('step: %d,loss_0:%f,loss_1:%f,acc:%f'%(count,tra_loss_0,tra_loss_1,this_acc))
            if count % int(steps_per_epoch*3) == 0 and count >= 0.8*total_epochs*steps_per_epoch:      
                checkpoint_path = os.path.join(logs_train_dir, 'model.ckpt')
                saver.save(sess, checkpoint_path, global_step=count)
            count += 1
                
        except tf.errors.OutOfRangeError:
            print('Done!')
            break